"""
Base Payment processor used by all derived processors.
"""
import django.dispatch

from copy import deepcopy
from datetime import timedelta, date
from calendar import mdays
from django.db.models import Sum
from django.conf import settings
from django.utils import timezone
from vendor import config
from vendor.models import Payment, Invoice, Receipt, Address
from vendor.models.choice import PurchaseStatus, TermType
##########
# SIGNALS

vendor_pre_authorization = django.dispatch.Signal()
vendor_process_payment =  django.dispatch.Signal()
vendor_post_authorization =  django.dispatch.Signal()

#############
# BASE CLASS

class PaymentProcessorBase(object):
    """
    Setup the core functionality for all processors.
    """
    API_ENDPOINT = None

    status = None
    invoice = None
    provider = None
    payment = None
    payment_info = {}
    billing_address = {}
    transaction_token = None
    transaction_id = ""
    transaction_submitted = False
    transaction_message = {}
    transaction_response = {}


    def __init__(self, invoice):
        """
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        self.set_invoice(invoice)
        self.provider = self.__class__.__name__
        self.processor_setup()
        self.set_api_endpoint()

    def set_api_endpoint(self):
        """
        Sets the API endpoint for debugging or production.It is dependent on the VENDOR_STATE
        enviornment variable. Default value is DEBUG for the VENDOR_STATE this function
        should be overwrote upon necesity of each Payment Processor
        """
        if config.VENDOR_STATE == 'DEBUG':
            self.API_ENDPOINT = None
        elif config.VENDOR_STATE == 'PRODUCTION':
            self.API_ENDPOINT = None

    def processor_setup(self):
        """
        This is for setting up any of the settings needed for the payment processing.
        For example, here you would set the 
        """
        pass

    def set_payment_info(self, **kwargs):
        self.payment_info = kwargs

    def set_invoice(self, invoice):
        self.invoice = invoice

    def create_payment_model(self):
        """
        Create payment instance with base information to track payment submissions
        """
        self.payment = Payment(profile=self.invoice.profile,
                               amount=self.invoice.total,
                               provider=self.provider,
                               invoice=self.invoice,
                               created=timezone.now()
                               )
        self.payment.result['account_number'] = self.payment_info.cleaned_data.get('card_number')[-4:]
        self.payment.payee_full_name = self.payment_info.cleaned_data.get('full_name')
        self.payment.payee_company = self.billing_address.cleaned_data.get('company')

        billing_address = self.billing_address.save(commit=False)
        billing_address, created = self.invoice.profile.get_or_create_address(billing_address)
        if created:
            billing_address.profile = self.invoice.profile
            billing_address.save()

        self.payment.billing_address = billing_address
        self.payment.save()

    def save_payment_transaction_result(self, payment_success, transaction_id, result_info):
        """
        Saves the result output of any transaction. 
        """
        self.payment.success = payment_success
        self.payment.transaction = transaction_id
        self.payment.result.update(result_info)
        self.payment.save()

    def update_invoice_status(self, new_status):
        """
        Updates the Invoice status if the transaction was submitted.
        Otherwise it returns the invoice to the Cart. The error is saved in 
        the payment for the transaction.
        """
        if self.transaction_submitted:
            self.invoice.status = new_status
        else:
            self.invoice.status = Invoice.InvoiceStatus.CART
        self.invoice.save()

    def is_payment_and_invoice_complete(self):
        """
        If payment was successful and invoice status is complete returns True. Otherwise
        false and no receipts should be created.
        """
        if self.payment.success and self.invoice.status == Invoice.InvoiceStatus.COMPLETE:
            return True
        return False
    
    def get_future_date_months(self, today, add_months):
        """
        Returns a datetime object with the a new added months
        """
        newday = today.day
        newmonth = (((today.month - 1) + add_months) % 12) + 1
        newyear  = today.year + (((today.month - 1) + add_months) // 12)
        if newday > mdays[newmonth]:
            newday = mdays[newmonth]
        if newyear % 4 == 0 and newmonth == 2:
            newday += 1
        return date(newyear, newmonth, newday)

    def get_future_date_days(self, today, add_days):
        """
        Returns a datetime object with the a new added days
        """
        return today + timedelta(days=add_days)

    def create_receipt_by_term_type(self, product, order_item, term_type):
        today = timezone.now()
        receipt = Receipt()
        receipt.profile = self.invoice.profile
        receipt.order_item = order_item
        receipt.transaction = self.payment.transaction
        receipt.status = PurchaseStatus.COMPLETE
        receipt.start_date = today
        if term_type == TermType.PERPETUAL or term_type == TermType.ONE_TIME_USE:
            receipt.auto_renew = False
        elif term_type == TermType.SUBSCRIPTION:
            total_months = int(order_item.offer.term_details['period_length']) * int(order_item.offer.term_details['payment_occurrences'])
            receipt.end_date = self.get_future_date_months(today, total_months)
            receipt.auto_renew = True
        else:
            total_months = term_type - 100
            receipt.end_date = self.get_future_date_months(today, total_months)
            receipt.auto_renew = True

        return receipt

    def create_order_item_receipt(self, order_item):
        """
        Creates a receipt for every product in the order item according to its,
        offering term type. 
        """
        for product in order_item.offer.products.all():
            receipt = self.create_receipt_by_term_type(product, order_item, order_item.offer.terms)
            receipt.save()
            receipt.products.add(product)

    def create_receipts(self, order_items):
        """
        It then creates receipt for the order items supplied. 
        """
        for order_item in order_items.all():
            self.create_order_item_receipt(order_item)

    def update_subscription_receipt(self, subscription, subscription_id, status):
        """
        subscription: OrderItem
        subscription_id: int
        status: PurchaseStatus
        """
        subscription_receipt = self.invoice.order_items.get(offer=subscription.offer).receipts.get(transaction=self.payment.transaction)
        subscription_receipt.meta['subscription_id'] = subscription_id
        subscription_receipt.status = status
        subscription_receipt.save()

    def amount(self):   # Retrieves the total amount from the invoice
        self.invoice.update_totals()
        return self.invoice.total

    def amount_without_subscriptions(self):
        subscription_total = sum([ oi.total for oi in self.invoice.order_items.filter(offer__terms=TermType.SUBSCRIPTION)])

        amount = self.invoice.total - subscription_total
        return amount

    def get_transaction_id(self):
        return "{}-{}-{}-{}".format(self.invoice.profile.pk, settings.SITE_ID, self.invoice.pk, str(self.payment.created)[-12:-6])

    def get_billing_address_form_data(self, form_data, form_class):
        self.billing_address = form_class(form_data)
    
    def get_payment_info_form_data(self, form_data, form_class):
        self.payment_info = form_class(form_data)

    def is_data_valid(self):
        if not (self.billing_address.is_valid() and self.payment_info.is_valid() and self.invoice and self.invoice.order_items.count()):
            return False
        return True
    #-------------------
    # Data for the View

    def get_checkout_context(self, request=None, context={}):
        '''
        The Invoice plus any additional values to include in the payment record.
        '''
        # context = deepcopy(context)
        context['invoice'] = self.invoice
        return context

    def get_header_javascript(self):
        """
        Scripts that are expected to show in the top of the template.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_javascript(self):
        """
        Scripts added to the bottom of the page in the normal js location.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_template(self):
        """
        Unique partial template for the processor
        """
        pass

    #-------------------
    # Process a Payment
    def authorize_payment(self):
        """
        This runs the chain of events in a transaction.
        
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        # TODO: Should this validation be outside the call to authorize the payment the call?
        # Why bother to call the processor is the forms are wrong
        if not self.is_data_valid():
            return None

        self.status = PurchaseStatus.QUEUED     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_pre_authorization.send(sender=self.__class__, invoice=self.invoice)

        self.pre_authorization()

        self.status = PurchaseStatus.ACTIVE     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_process_payment.send(sender=self.__class__, invoice=self.invoice)

        if not self.invoice.total:
            self.free_payment()
        elif self.invoice.get_one_time_transaction_order_items():
            self.create_payment_model()
            self.process_payment()
            self.save_payment_transaction_result(self.transaction_submitted, self.transaction_id, self.transaction_response)
            self.update_invoice_status(Invoice.InvoiceStatus.COMPLETE)
            if self.is_payment_and_invoice_complete():
                self.create_receipts(self.invoice.get_one_time_transaction_order_items())

        if self.invoice.get_recurring_order_items():
            self.process_subscriptions()        

        vendor_post_authorization.send(sender=self.__class__, invoice=self.invoice)
        self.post_authorization()

        #TODO: Set the status based on the result from the process_payment()

    def pre_authorization(self):
        """
        Called before the authorization begins.
        """
        pass

    def process_payment(self):
        """
        Called to handle the authorization.  
        This is where the core of the payment processing happens.
        """
        # Gateway Transaction goes here...
            
    def free_payment(self):
        """
        Called to handle an invoice with total zero.  
        This are the base internal steps to process a free payment.
        """
        self.create_payment_model()
        self.transaction_submitted = True

        self.payment.success = True
        self.payment.transation = f"{self.payment.pk}-free"
        self.payment.payee_full_name = " ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
        self.payment.save()
        
        self.update_invoice_status(Invoice.InvoiceStatus.COMPLETE)

        self.create_receipts(self.invoice.order_items.all())

    def post_authorization(self):
        """
        Called after the authorization is complete.
        """
        pass

    def capture_payment(self):
        """
        Called to handle the capture.  (some gateways handle this at the same time as authorize_payment() )
        """
        pass

    def void_payment(self):
        """
        Call to handle a payment that has not been settled and wants to be voided
        """
        pass

    #-------------------
    # Process a Subscription
    def process_subscriptions(self):
        """
        Process/subscribies recurring payments throught the payement gateway and creates a payment model for each subscription.
        If a payment is completed it will create a receipt for the subscription
        """
        if not self.is_card_valid():
            return None

        for subscription in self.invoice.get_recurring_order_items():
            self.create_payment_model()
            self.subscription_payment(subscription)
            self.save_payment_transaction_result(self.transaction_submitted, self.transaction_id, self.transaction_response)
            self.update_invoice_status(Invoice.InvoiceStatus.COMPLETE)
            if self.is_payment_and_invoice_complete():
                self.create_order_item_receipt(subscription)
        
    def subscription_payment(self, subscription):
        """
        Call handels the authrization and creation for a subscription.
        """
        # Gateway Transaction goes here...

    def subscription_info(self):
        pass

    def subscription_update_payment(self):
        pass

    def subscription_cancel(self):
        pass

    def is_card_valid(self):
        """
        Function to validate a credit card by method of makeing a microtransaction and voiding it if authorized.
        """
        pass

    #-------------------
    # Refund a Payment

    def refund_payment(self):
        pass

    


# App Settings
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

VENDOR_PRODUCT_MODEL = getattr(settings, "VENDOR_PRODUCT_MODEL", "vendor.Product")

VENDOR_PAYMENT_PROCESSOR = getattr(settings, "VENDOR_PAYMENT_PROCESSOR", "dummy.DummyProcessor")

DEFAULT_CURRENCY = getattr(settings, "DEFAULT_CURRENCY", "usd")

AVAILABLE_CURRENCIES = getattr(settings, "AVAILABLE_CURRENCIES", {'usd': _('USD Dollars')})

VENDOR_STATE = getattr(settings, "VENDOR_STATE", "DEBUG")

# Encryption settings
VENDOR_DATA_ENCODER = getattr(settings, "VENDOR_DATA_ENCODER", "vendor.encrypt.cleartext")



# Generated by Django 3.1 on 2020-09-15 18:39

from django.db import migrations, models
import django.db.models.deletion
import vendor.models.utils


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('vendor', '0002_auto_20200912_0142'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customerprofile',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='customer_profile', to='sites.site'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='invoices', to='sites.site'),
        ),
        migrations.AlterField(
            model_name='offer',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='product_offers', to='sites.site'),
        ),
    ]



# Generated by Django 3.1.3 on 2020-12-03 19:51
import uuid
from django.db import migrations


def generate_payment_uuid(apps, schema_editor):
    PaymentModel = apps.get_model('vendor', 'Payment')
    for payment in PaymentModel.objects.all():
        payment.uuid = uuid.uuid4()
        payment.save(update_fields=['uuid'])
    
def generate_receipt_uuid(apps, schema_editor):
    ReceiptModel = apps.get_model('vendor', 'Receipt')
    for receipt in ReceiptModel.objects.all():
        receipt.uuid = uuid.uuid4()
        receipt.save(update_fields=['uuid'])


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0014_term_types_update_value'),
    ]

    operations = [
        migrations.RunPython(generate_payment_uuid, reverse_code=migrations.RunPython.noop),
        migrations.RunPython(generate_receipt_uuid, reverse_code=migrations.RunPython.noop),
    ]



from django.utils import timezone
from django.db.models import F
from django.shortcuts import get_object_or_404, render, redirect
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse, reverse_lazy
from django.conf import settings
from django.utils.translation import ugettext as _
from django.core.exceptions import ObjectDoesNotExist

from django.views import View
from django.views.generic.edit import CreateView, UpdateView, DeleteView, FormView
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic import TemplateView
from django.http import HttpResponse

from vendor.models import Offer, OrderItem, Invoice, Payment, Address
from vendor.models.address import Address as GoogleAddress
from vendor.processors import PaymentProcessor
from vendor.forms import BillingAddressForm, CreditCardForm

from .vendor_admin import AdminDashboardView, AdminInvoiceDetailView, AdminInvoiceListView


payment_processor = PaymentProcessor              # The Payment Processor configured in settings.py

# class CartView(LoginRequiredMixin, DetailView):
#     '''
#     View items in the cart
#     '''
#     model = Invoice

#     def get_object(self):
#         profile, created = self.request.user.customer_profile.get_or_create(site=settings.SITE_ID)
#         return profile.get_cart()


# class AddToCartView(LoginRequiredMixin, TemplateView):
#     '''
#     Create an order item and add it to the order
#     '''

#     def get(self, *args, **kwargs):         # TODO: Move to POST
#         offer = Offer.objects.get(slug=self.kwargs["slug"])
#         profile, created = self.request.user.customer_profile.get_or_create(site=settings.SITE_ID)      # Make sure they have a cart

#         cart = profile.get_cart()
#         cart.add_offer(offer)

#         messages.info(self.request, _("Added item to cart."))

#         return redirect('vendor:cart')      # Redirect to cart on success


# class RemoveFromCartView(LoginRequiredMixin, DeleteView):
#     '''
#     Reduce the count of items from the cart and delete the order item if you reach 0
#     TODO: Change to form/POST for better security & flexibility
#     '''
#     def get(self, *args, **kwargs):         # TODO: Move to POST
#         offer = Offer.objects.get(slug=self.kwargs["slug"])

#         profile = self.request.user.customer_profile.get(site=settings.SITE_ID)      # Make sure they have a cart
#         cart = profile.get_cart()
#         cart.remove_offer(offer)

#         messages.info(self.request, _("Removed item from cart."))

#         return redirect('vendor:cart')      # Redirect to cart on success


# class PaymentView(LoginRequiredMixin, TemplateView):
#     '''
#     Review items and submit Payment
#     '''
#     template_name = "vendor/checkout.html"
#     # billing_address_form_class = BillingAddressForm
#     # card_form_class = CreditCardForm
#     # payment_processor = PaymentProcessor

#     # def get_context_data(self, **kwargs):
#     #     context = super().get_context_data(**kwargs)

#     #     profile = self.request.user.customer_profile.get(site=settings.SITE_ID) 
#     #     invoice = profile.invoices.get(status=Invoice.InvoiceStatus.CART)

#     #     processor = self.payment_processor(invoice)

#     #     context = processor.get_checkout_context(context=context)

#     #     context['billing_form'] = self.billing_form_class()
#     #     # TODO: Set below in the PaymentProcessor Context. It should set the address form and card form?
#     #     context['address_form'] = self.address_form_class(prefix='addr')
#     #     context['card_form'] = self.card_form_class(prefix='card')

#     #     return context

#     def get(self, request, *args, **kwargs):
#         context = super().get_context_data(**kwargs)
#         profile = self.request.user.customer_profile.get(site=settings.SITE_ID) 
#         invoice = profile.invoices.get(status=Invoice.InvoiceStatus.CART)

#         processor = payment_processor(invoice)

#         context = processor.get_checkout_context(context=context)

#         return render(request, self.template_name, context)


#     def post(self, request, *args, **kwargs):
#         context = super().get_context_data(**kwargs)
#         profile = request.user.customer_profile.get(site=settings.SITE_ID) 
#         invoice = Invoice.objects.get(profile=profile, status=Invoice.InvoiceStatus.CART)
                
#         credit_card_form = CreditCardForm(request.POST, prefix='credit-card')
#         billing_address_form = BillingAddressForm(request.POST, prefix='billing-address')
        
#         if not (billing_address_form.is_valid() or credit_card_form.is_valid()):
#             return render(request, self.template_name, processor.get_checkout_context(context))
        
#         processor = payment_processor(invoice)

#         processor.process_payment(request)
#         if processor.transaction_submitted:
#             return redirect('vendor:purchase-summary', pk=invoice.pk)   # redirect to the summary page for the above invoice
#             # TODO: invoices should have a UUID attached to them
#             # return redirect('vendor:purchase-summary', pk=processor.invoice.payments.filter(success=True).values_list('pk'))    # TODO: broken
#         else:
#             return render(request, self.template_name, processor.get_checkout_context(request, context))

        
# class PaymentView(LoginRequiredMixin, DetailView):
#     model = Payment


######################
# Order History Views

class OrderHistoryListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated by the current user on the current site.
    '''
    model = Invoice
    #TODO: filter to only include the current user's order history

    def get_queryset(self):
        try:
            return self.request.user.customer_profile.get().invoices.filter(status__gt=Invoice.InvoiceStatus.CART)  # The profile and user are site specific so this should only return what's on the site for that user excluding the cart
        except ObjectDoesNotExist:         # Catch the actual error for the exception
            return []   # Return empty list if there is no customer_profile

class OrderHistoryDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated by the current user on the current site.
    '''
    template_name = "vendor/invoice_history_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


# class PaymentView(LoginRequiredMixin, TemplateView):
# class PaymentView(TemplateView):
#     '''
#     Review items and submit Payment
#     '''
#     template_name = "vendor/checkout.html"

#     # GET returns the invoice with the items and the estimated totals.

#     def get(self, request, *args, **kwargs):
#         profile = request.user.customer_profile.get(site=settings.SITE_ID)      # Make sure they have a cart
#         order = Invoice.objects.get(profile=profile, status=0)

#         ctx = payment_processor.get_checkout_context(order, customer_id=str(request.user.pk))
#         print(ctx)

#         return render(request, self.template_name, ctx)


#     def post(self, request, *args, **kwargs):
#         print(request)
#         print(request.POST)
#         print(request.headers)
#         profile = request.user.customer_profile.get(site=settings.SITE_ID)      # Make sure they have a cart
#         order = Invoice.objects.get(profile=profile, status=0)
#         token = request.POST.get("stripeToken")

#         # amount= int(order.total * 100)
#         # currency = order.currency

#         # description = "Invoice #{} for ... in the amount of {}".format(order.pk, amount)

#         # # stripe.Charge.create(
#         # #     amount=amount,
#         # #     currency=currency,
#         # #     source=token,
#         # #     description=description,
#         # # )

#         # order.status = 20




# class NewAddToCartView(CreateView):
#     model = OrderItem
#     form_class = AddToCartModelForm                                                         # todo: Move to a regular Form from a ModelForm since the advantages of the ModelForm are not used...
#     success_url = reverse_lazy('vendor-user-cart-retrieve')

#     def form_valid(self, form):

#         invoice, invoice_created = self.request.user.invoice_set.get_or_create(status=OrderStatus.CART.value)   # todo: Need to see if there is a way to make sure there is a cart and handle the theoretical case of multiples (thought it should be a 'singleton').

#         quantity = form.instance.quantity                                                               # How many?
#         offer = Offer.objects.get( sku=self.kwargs['sku'] )                                             # SKU is in the URL...

#         order_item, created = invoice.order_items.get_or_create(offer=offer)                            # Get or Create an order item with the matching values

#         if not created:                                                                                 # Add the quantity if it already exists
#             order_item.quantity += quantity
#         else:                                                                                           # Set the value if it does not
#             order_item.quantity = quantity

#         order_item.save()
#         messages.info(self.request, "\"{}\"".format( order_item.product_name ) + _(" added to cart."))

#         return redirect(self.success_url)


# class AddToCartView(LoginRequiredMixin, CreateView):
#     model = OrderItem
#     form_class = AddToCartModelForm
#     success_url = reverse_lazy('vendor-user-cart-retrieve')

#     # def get_object(self, queryset=None):
#     #     self.offer = self.request.POST.get('offer')         # todo: This is meant to return a model object, not a string...
#     #     return self.offer

#     def form_valid(self,form):
#         # offer = Offer.objects.get(id = self.get_object()) 
#         offer = Offer.objects.get(sku=self.request.POST.get('offer') ) 
    
#         price = offer.current_price() #sale_price.filter(start_date__lte= timezone.now(), end_date__gte=timezone.now()).order_by('priority').first()

#         invoice = Invoice.objects.filter(user = self.request.user, status=OrderStatus.CART.value)

#         # check if there is an invoice with status= cart for the user
#         if invoice.exists():

#             invoice_qs = invoice[0]

#             order_item = invoice_qs.order_items.filter(offer__sku = offer.sku)

#             # check if the order_item is there in the invoice, if yes increase the quantity
#             if order_item.exists():
#                 order_item.update(quantity=F('quantity')+1)
#                 messages.info(self.request, _("Item quantity was updated."))
#                 return redirect(self.success_url)

#             # Create an order_item object for the invoice
#             else:
#                 order_item = self.model.objects.create(
#                     invoice = invoice_qs,
#                     offer=offer,
#                     price=price
#                 )
#                 messages.info(self.request, _("Item added to cart."))
#                 return redirect(self.success_url)

#         # Create a invoice object with status = cart
#         else:
#             ordered_date = timezone.now()
#             invoice = Invoice.objects.create(user=self.request.user, ordered_date=ordered_date)

#             order_item = OrderItem.objects.create(invoice = invoice, offer = offer, price = price)
#             messages.info(self.request, _("Item added to cart."))
#             return redirect(self.success_url)


# class RetrieveCartView(LoginRequiredMixin, ListView):
#     model = Invoice

#     def get_queryset(self):
#         invoice = self.model.objects.filter(user = self.request.user, status=OrderStatus.CART.value).first()
#         if invoice:
#             return invoice.order_items.all()

#         return None

#     def get_context_data(self, **kwargs):
#         context = super().get_context_data(**kwargs)
#         count = 0
#         invoice = self.model.objects.filter(user = self.request.user, status=OrderStatus.CART.value)
#         if invoice:
#             context['invoice'] = True

#         if self.get_queryset():
#             for item in self.get_queryset():
#                 count += item.quantity

#             context['item_count'] = count
        
#         return context

# class RemoveFromCartView(LoginRequiredMixin, DeleteView):
#     model = OrderItem
#     success_url = reverse_lazy('vendor-user-cart-retrieve')

#     def get_object(self, queryset=None):
#         invoice = Invoice.objects.get(user = self.request.user, status=OrderStatus.CART.value)
#         return invoice.order_items.get(offer__sku = self.kwargs['sku'])

#     def form_valid(self, form):
#         self.get_object().delete()
#         messages.info(self.request, _("Item removed from cart"))
#         return redirect(self.success_url)


# class CartItemQuantityEditView(LoginRequiredMixin, View):
  
#     def post(self, request, *args, **kwargs):

#         quantity = request.POST.get('quantity')

#         order_item = OrderItem.objects.filter(id = self.kwargs['id']).update(quantity = quantity )  

#         return redirect(reverse_lazy('vendor-user-cart-retrieve'))

         
# class DeleteCartView(LoginRequiredMixin, DeleteView):
#     model = Invoice

#     def delete(request):

#         invoice = self.model.filter(user = request.user, status = 0)

#         if invoice:
#             invoice[0].order_items.all().delete()
#             invoice[0].delete()

#             messages.info(request, "User Cart deleted.")
#             return redirect("vendor:vendor_index")

#         else:
#             messages.info(request, "You do not have an active cart.")
#             return redirect("vendor:vendor_index")


# class RetrieveOrderSummaryView(LoginRequiredMixin, ListView):
#     model = Invoice
#     template_name = "vendor/ordersummary.html"

#     def get_queryset(self):
#         invoice = self.model.objects.get(user = self.request.user, status=OrderStatus.CART.value)
#         return invoice.order_items.all()

#     def get_context_data(self, **kwargs):
#         context = super().get_context_data(**kwargs)
#         count = 0
#         total = 0
    
#         for item in self.get_queryset():
#             count +=  item.quantity
#             total += item.total

#         context['item_count'] = count
#         context['order_total'] = total
#         context['amount'] = total * 100
#         context['key'] = settings.STRIPE_TEST_PUBLIC_KEY
        
#         return context


# class PaymentProcessingView(LoginRequiredMixin, View):

#     def post(self, *args, **kwargs):
#         stripe.api_key = settings.STRIPE_TEST_SECRET_KEY
#         token = self.request.POST['stripeToken']

#         invoice = Invoice.objects.get(user = self.request.user, status=OrderStatus.CART.value)

#         total = 0
        
#         order_items = invoice.order_items.all()

#         for item in order_items:
#             total += item.total

#         try:
#             customer_profile = CustomerProfile.objects.get(user = self.request.user)
        
#         except:
#             customer_profile = CustomerProfile.objects.create(user = self.request.user, currency = 'usd', attrs = {})
        
#         stripe_customer_id = customer_profile.attrs.get('stripe_customer_id', None)

#         if stripe_customer_id != '' and stripe_customer_id is not None:
#                     customer = stripe.Customer.retrieve(
#                         stripe_customer_id)

#         else:
#             customer = stripe.Customer.create(
#                 email=self.request.user.email,
#                 card=token
#             )
#             customer_profile.attrs['stripe_customer_id'] = customer['id']
#             customer_profile.save()
     
#         charge = stripe.Charge.create(
#             amount=int(total * 100),
#             currency='usd',
#             description='TrainingCamp Charge',      # todo: this need to be set relative to the offer item.
#             customer=customer.id,
#             metadata={'invoice_id': invoice.id},
#         )

#         invoice.status = OrderStatus.COMPLETE.value
#         invoice.attrs = {'charge': charge.id}
#         invoice.ordered_date = timezone.now()
#         invoice.save()

#         for items in order_items:
#             Purchase.objects.create(order_item = items, product = items.offer.product, user = self.request.user)
        
#         messages.success(self.request, _("Your order was successful!"))
#         return redirect(reverse_lazy('vendor-user-order-retrieve', kwargs = {'id': invoice.id}))


# class RetrieveOrderView(LoginRequiredMixin, ListView):
#     model = Purchase

#     def get_queryset(self):
#         return self.model.objects.filter(order_item__invoice__id = self.kwargs['id'])


# class RetrievePurchaseView(LoginRequiredMixin, DetailView):
#     model = Purchase

#     def get_object(self):
#         return self.model.objects.get(id=self.kwargs['id'])

#     def get_context_data(self, **kwargs):
#         context = super().get_context_data(**kwargs)

#         status = self.get_object().status

#         if status is PurchaseStatus.ACTIVE.value or status is PurchaseStatus.QUEUED.value:
#             context['refund'] = "Request Refund"
#             context['active'] = True

#         elif status is PurchaseStatus.CANCELED.value:
#             context['refund'] = "Refund Requested"

#         elif status is PurchaseStatus.REFUNDED.value:
#             context['refund'] = "Refund Issued"

#         return context


# class RetrievePurchaseListView(LoginRequiredMixin, ListView):
#     model = Purchase 

#     def get_queryset(self):
#         return self.request.user.purchase_set.all()

       
# class RequestRefundView(LoginRequiredMixin, CreateView):
#     model = Refund
#     form_class = RequestRefundForm

#     def get_object(self, queryset=None):
#         self.purchase = Purchase.objects.get(id=self.kwargs['id'])
#         return self.purchase

#     def form_valid(self,form):

#         purchase = self.get_object()

#         reason = self.request.POST.get('reason')

#         Refund.objects.create(purchase = purchase, reason = reason, user = self.request.user)

#         purchase.status = PurchaseStatus.CANCELED.value
#         purchase.save()

#         messages.info(self.request, _("Refund request created"))
#         return redirect(reverse_lazy('vendor-user-purchase-list'))


# # class RetrieveRefundRequestsView(LoginRequiredMixin, ListView):
# #     model = Refund 

# #     def get_queryset(self):
# #         return self.model.objects.all()


# class IssueRefundView(LoginRequiredMixin, CreateView):
#     model = Refund 
#     success_url = reverse_lazy('vendor-retrieve-refund-requests')

#     def get_queryset(self):
#         return self.model.objects.get(id = self.kwargs['id'])

#     def post(self, request, *args, **kwargs):
#         stripe.api_key = settings.STRIPE_TEST_SECRET_KEY
        
#         refund = self.get_queryset()
#         purchase = refund.purchase

#         invoice = purchase.order_item.invoice 

#         charge = invoice.attrs['charge']

#         stripe.Refund.create(
#             charge=charge, amount = int(purchase.order_item.price.cost * 100))

#         refund.accepted = True
#         refund.save()

#         purchase.status = PurchaseStatus.REFUNDED.value
#         purchase.save()

#         messages.info(request, _("Refund was issued"))

#         return redirect(reverse_lazy('vendor-retrieve-refund-requests'))





   

  



# Generated by Django 3.1.3 on 2020-12-03 19:33

from django.db import migrations

from vendor.models.choice import TermType

def term_type_update_value(apps, schema_editor):
    OfferModel = apps.get_model('vendor', 'Offer')

    for offer in OfferModel.objects.filter(terms__lt=100):
        if offer.terms == 0:
            offer.terms = TermType.PERPETUAL
        elif offer.terms == 10:
            offer.terms = TermType.SUBSCRIPTION
        elif offer.terms == 11:
            offer.terms = TermType.MONTHLY_SUBSCRIPTION
        elif offer.terms == 12:
            offer.terms = TermType.QUARTERLY_SUBSCRIPTION
        elif offer.terms == 13:
            offer.terms = TermType.SEMIANNUAL_SUBSCRIPTION
        elif offer.terms == 14:
            offer.terms = TermType.ANNUAL_SUBSCRIPTION
        elif offer.terms == 20:
            offer.terms = TermType.ONE_TIME_USE
        offer.save()

class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0013_auto_20201202_1759'),
    ]

    operations = [
        migrations.RunPython(term_type_update_value, reverse_code=migrations.RunPython.noop),
    ]



from django.apps import apps
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView
from django.views.generic.list import ListView
from vendor.config import VENDOR_PRODUCT_MODEL
from vendor.models import Invoice, Offer, Price
from vendor.forms import ProductForm, OfferForm, PriceForm, PriceFormSet
from django.utils.translation import ugettext as _

Product = apps.get_model(VENDOR_PRODUCT_MODEL)
#############
# Admin Views

class AdminDashboardView(LoginRequiredMixin, ListView):
    '''
    List of the most recent invoices generated on the current site.
    '''
    template_name = "vendor/manage/dashboard.html"
    model = Invoice

    def get_queryset(self):
        return self.model.on_site.all()[:10]    # Return the most recent 10


class AdminInvoiceListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated on the current site.
    '''
    template_name = "vendor/manage/invoice_list.html"
    model = Invoice

    def get_queryset(self):
        return self.model.on_site.filter(status__gt=Invoice.InvoiceStatus.CART).order_by('updated')  # ignore cart state invoices


class AdminInvoiceDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/invoice_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class AdminProductListView(LoginRequiredMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/products.html"
    model = Product
    queryset = Product.on_site.all()


class AdminProductUpdateView(LoginRequiredMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/product.html"
    model = Product
    fields = ['sku', 'name', 'site', 'available', 'description', 'meta']
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'
    success_url = reverse_lazy('vendor_admin:manager-product-list')


class AdminProductCreateView(LoginRequiredMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/product.html"
    model = Product
    fields = ['sku', 'name', 'site', 'available', 'description', 'meta']
    success_url = reverse_lazy('vendor_admin:manager-product-list')


class AdminOfferListView(LoginRequiredMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offers.html"
    model = Offer
    queryset = Offer.on_site.all()


class AdminOfferUpdateView(LoginRequiredMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'
    template_name_suffix = '_update_form'

    def get_initial(self):
        return {'products': self.object.products.all()}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['formset'] = PriceFormSet(instance=self.object)

        return context


    def form_valid(self, form):
        price_formset = PriceFormSet(self.request.POST, self.request.FILES, instance=Offer.objects.get(uuid=self.kwargs['uuid']))

        if not (price_formset.is_valid() or form.is_valid()):
            return render(self.request, self.template_name, {'form': form, 'formsert': price_formset})

        offer = form.save(commit=False)

        if len(form.cleaned_data['products']) > 1:
            offer.bundle=True
        
        offer.save()

        for product in form.cleaned_data['products']:
            offer.products.add(product)

        for price_form in price_formset:
            price = price_form.save(commit=False)
            price.offer = offer
            price.save()

        return redirect('vendor_admin:manager-offer-list')


class AdminOfferCreateView(LoginRequiredMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['formset'] = PriceFormSet()

        return context

    def post(self, request):
        offer_form = self.form_class(request.POST)
        price_formset = PriceFormSet(request.POST)


        if not (price_formset.is_valid() and offer_form.is_valid()):
            return render(request, self.template_name, {'form': offer_form, 'formset': price_formset})
        
        product_currencies = {}
        price_currency = [ price_form.cleaned_data['currency'] for price_form in price_formset ]

        for product in Product.objects.filter(pk__in=offer_form.cleaned_data['products']):
            for currency in product.meta['msrp'].keys():
                product_currencies[currency] = currency

        for price_form in price_formset:
            if price_form.cleaned_data['currency'] not in product_currencies:
                price_formset[0].add_error('currency', _('Invalid currency'))
                return render(request, self.template_name, {'form': offer_form, 'formset': price_formset})
                

        offer = offer_form.save(commit=False)
        if len(offer_form.cleaned_data['products']) > 1:
            offer.bundle=True

        offer.save()
        for product in offer_form.cleaned_data['products']:
            offer.products.add(product)


        for price_form in price_formset:
            price = price_form.save(commit=False)
            price.offer = offer
            price.save()

        return redirect('vendor_admin:manager-offer-list')


import uuid
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .base import CreateUpdateModelBase
from vendor.models.choice import PurchaseStatus

from vendor.config import VENDOR_PRODUCT_MODEL

class Receipt(CreateUpdateModelBase):
    '''
    A link for all the purchases a user has made. Contains subscription start and end date.
    This is generated for each item a user purchases so it can be checked in other code.
    '''
    uuid = models.UUIDField(_("UUID"), editable=False, unique=True, default=uuid.uuid4, null=False, blank=False)
    profile = models.ForeignKey("vendor.CustomerProfile", verbose_name=_("Purchase Profile"), null=True, on_delete=models.CASCADE, related_name="receipts")
    order_item = models.ForeignKey('vendor.OrderItem', verbose_name=_("Order Item"), on_delete=models.CASCADE, related_name="receipts")
    start_date = models.DateTimeField(_("Start Date"), blank=True, null=True)
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True)
    auto_renew = models.BooleanField(_("Auto Renew"), default=False)        # For subscriptions
    vendor_notes = models.JSONField(_("Vendor Notes"), default=dict)
    transaction = models.CharField(_("Transaction"), max_length=80)
    status = models.IntegerField(_("Status"), choices=PurchaseStatus.choices, default=0)       # Fulfilled, Refund
    meta = models.JSONField(_("Meta"), default=dict)
    # TODO: Add final purchase price to the receipt for tracking.
    # TODO: Add Site field for easier tracking?
    # the product connection comes from the ProductModelBase to not trigger a migration on subclassing PMB

    class Meta:
        verbose_name = "Receipt"
        verbose_name_plural = "Receipts"

    def __str__(self):
        return "%s - %s - %s" % (self.profile.user.username, self.order_item.offer.name, self.created.strftime('%Y-%m-%d %H:%M'))

    def get_absolute_url(self):
        return reverse('vendor:customer-receipt', kwargs={'uuid': self.uuid})






# TODO: Need to think of a better way to check if a user has product in a mixin
# from django.apps import apps
# from django.conf import settings
# from .config import VENDOR_PRODUCT_MODEL
# Product = apps.get_model(VENDOR_PRODUCT_MODEL)

# class UserOwnsProductMixin(object):
#     """Verify if user has product to view"""
#     def get_context_data(self, **kwargs):
#         context = super().get_context_data(**kwargs)
#         product = Product.objects.get(slug=self.kwargs["slug"])
        
#         context['owns_product'] = self.request.user.customer_profile.get(site=settings.SITE_ID).has_product(product)
        
#         return context


from calendar import monthrange
from datetime import datetime
from django import forms
from django.apps import apps
from django.contrib.auth import get_user_model
from django.db.models import IntegerChoices
from django.forms import inlineformset_factory
from django.forms.widgets import SelectDateWidget
from django.utils.translation import ugettext_lazy as _

from .config import VENDOR_PRODUCT_MODEL
from .models import Address, Offer, OrderItem, Price
from .models.choice import PaymentTypes, TermType

Product = apps.get_model(VENDOR_PRODUCT_MODEL)

        
class PriceForm(forms.ModelForm):
    start_date = forms.DateField(label=_("Start Date"), widget=SelectDateWidget())
    end_date = forms.DateField(label=_("End Date"), widget=SelectDateWidget())
    class Meta:
        model = Price
        fields = ['cost', 'currency', 'start_date', 'end_date', 'priority']


class ProductForm(forms.ModelForm):
    class Meta:
        model = Product
        fields = ['sku', 'name', 'site', 'available', 'description', 'meta']


class OfferForm(forms.ModelForm):
    products = forms.ModelMultipleChoiceField(label=_("Available Products:"), required=True, queryset=Product.on_site.filter(available=True))
    start_date = forms.DateField(label=_("Start Date"), widget=SelectDateWidget())
    end_date = forms.DateField(label=_("End Date"), widget=SelectDateWidget())
    term_start_date = forms.DateField(label=_("Term Start Date"), widget=SelectDateWidget())

    class Meta:
        model = Offer
        fields = ['name', 'start_date', 'end_date', 'terms', 'term_details', 'term_start_date', 'available', 'offer_description', 'allow_multiple']

    def clean(self):
        cleaned_data = super().clean()

        if not cleaned_data['name']:
            product_names = [ product.name for product in self.cleaned_data['products'] ]
            if len(product_names) == 1:
                self.cleaned_data['name'] = product_names[0]
            else:
                self.cleaned_data['name'] = _("Bundle: ") + ", ".join(product_names)

        if self.data['terms'] == TermType.SUBSCRIPTION and not cleaned_data['term_details']:
            self.add_error('term_details', _("Invalid term details for subscription"))

        return cleaned_data


PriceFormSet = inlineformset_factory(
    Offer,
    Price,
    form=PriceForm,
    can_delete=True,
    exclude=('offer',),
    validate_max=True,
    min_num=1,
    extra=0)


class AddressForm(forms.ModelForm):

    class Meta:
        model = Address
        fields = ['name', 'first_name', 'last_name', 'country', 'address_1', 'address_2', 'locality', 'state', 'postal_code']

    def __init__(self, *args, **kwargs):
        super(AddressForm, self).__init__(*args, **kwargs)
        self.fields['name'].hidden = True
        self.fields['address_1'].widget.attrs.update({'placeholder' : _('Enter Address')})
        self.fields['address_2'].widget.attrs.update({'placeholder' : _('Enter Apt, Suite, Unit, Building, Floor, etc')})
        self.fields['locality'].widget.attrs.update({'placeholder' : _('Enter City')})


class AccountInformationForm(AddressForm):
    email = forms.EmailField(label=_('Email Address'), required=True)

    class Meta:
        model = Address
        fields = ['name', 'first_name', 'last_name', 'email', 'country', 'address_1', 'address_2', 'locality', 'state', 'postal_code']


class BillingAddressForm(AddressForm):
    same_as_shipping = forms.BooleanField(label=_("Billing address is the same as shipping address"), required=False)

    class Meta:
        model = Address
        fields = ['same_as_shipping', 'name', 'first_name', 'last_name', 'country', 'address_1', 'address_2', 'locality', 'state', 'postal_code']

    def __init__(self, *args, **kwargs):
        super(BillingAddressForm, self).__init__(*args, **kwargs)
        self.fields['country'].label = _('Billing Country/Region')


class CreditCardField(forms.CharField):

    # validates almost all of the example cards from PayPal
    # https://www.paypalobjects.com/en_US/vhelp/paypalmanager_help/credit_card_numbers.htm
    cards = [
        {
            'type': 'maestro',
            'patterns': [5018, 502, 503, 506, 56, 58, 639, 6220, 67],
            'length': [12, 13, 14, 15, 16, 17, 18, 19],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'forbrugsforeningen',
            'patterns': [600],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'dankort',
            'patterns': [5019],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'visa',
            'patterns': [4],
            'length': [13, 16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'mastercard',
            'patterns': [51, 52, 53, 54, 55, 22, 23, 24, 25, 26, 27],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'amex',
            'patterns': [34, 37],
            'length': [15],
            'cvvLength': [3, 4],
            'luhn': True
        }, {
            'type': 'dinersclub',
            'patterns': [30, 36, 38, 39],
            'length': [14],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'discover',
            'patterns': [60, 64, 65, 622],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'unionpay',
            'patterns': [62, 88],
            'length': [16, 17, 18, 19],
            'cvvLength': [3],
            'luhn': False
        }, {
            'type': 'jcb',
            'patterns': [35],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }
    ]

    def __init__(self, placeholder=None, *args, **kwargs):
        super(CreditCardField, self).__init__(
            # override default widget
            widget=forms.widgets.TextInput(attrs={'placeholder': placeholder, 'type':'tel'}), *args, **kwargs)

    default_error_messages = {
        'invalid': _(u'The credit card number is invalid'),
    }

    def clean(self, value):

        # ensure no spaces or dashes
        value = value.replace(' ', '').replace('-', '')

        # get the card type and its specs
        card = self.card_from_number(value)

        # if no card found, invalid
        if not card:
            raise forms.ValidationError(self.error_messages['invalid'])

        # check the length
        if not len(value) in card['length']:
            raise forms.ValidationError(self.error_messages['invalid'])

        # test luhn if necessary
        if card['luhn']:
            if not self.validate_mod10(value):
                raise forms.ValidationError(self.error_messages['invalid'])

        return value

    def card_from_number(self, num):
        # find this card, based on the card number, in the defined set of cards
        for card in self.cards:
            for pattern in card['patterns']:
                if (str(pattern) == str(num)[:len(str(pattern))]):
                    return card

    def validate_mod10(self, num):
        # validate card number using the Luhn (mod 10) algorithm
        checksum, factor = 0, 1
        for x in reversed(num):
            for y in str(factor * int(x)):
                checksum += int(y)
            factor -= 3
            factor = abs(factor)
        return checksum % 10 == 0


class PaymentFrom(forms.Form):
    payment_type = forms.ChoiceField(label=_("Payment Type"), choices=PaymentTypes.choices, widget=forms.widgets.HiddenInput)


class CreditCardForm(PaymentFrom):
    full_name = forms.CharField(required=True, label=_("Name on Card"), max_length=80)
    card_number = CreditCardField(label=_("Credit Card Number"), placeholder=u'0000 0000 0000 0000', min_length=12, max_length=19)
    expire_month = forms.ChoiceField(required=True, label=_("Expiration Month"), choices=[(x, x) for x in range(1, 13)])
    expire_year = forms.ChoiceField(required=True, label=_("Expiration Year"))
    cvv_number = forms.CharField(required=True, label=_("CVV Number"), max_length=4, min_length=3, widget=forms.TextInput(attrs={'size': '4'}))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        today = datetime.now()
        self.fields['expire_year'].choices = [(x, x) for x in range(today.year, today.year + 15)]
        self.fields['expire_year'].initial = (today.year, today.year)


    def clean(self):
        cleaned_data = super(CreditCardForm, self).clean()
        expire_month = cleaned_data.get('expire_month')
        expire_year = cleaned_data.get('expire_year')

        if not self.expiration_date_valid(expire_month, expire_year):
            del(cleaned_data['expire_month'])
            del(cleaned_data['expire_year'])
             
        return cleaned_data

    def clean_expire_month(self):
        expire_month = self.cleaned_data.get('expire_month')

        if not expire_month:
            raise forms.ValidationError(_("You must select a valid expiration month"), code='required')

        return expire_month

    def clean_expire_year(self):
        expire_year = self.cleaned_data.get('expire_year')

        if not expire_year:
            raise forms.ValidationError(_("You must select a valid expiration year"), code='required')

        return expire_year
    
    def clean_cvv_number(self):
        cvv_number = self.cleaned_data.get('cvv_number')

        if not cvv_number.isdigit():
            raise forms.ValidationError(_("CVV must be all digits"))

        return cvv_number
    
    def expiration_date_valid(self, expire_month, expire_year):
        year = int(expire_year)
        month = int(expire_month)

        # find last day of the month
        day = monthrange(year, month)[1]
        expire = datetime(year, month, day)

        if datetime.now() > expire:
            self._errors["expire_year"] = self.error_class([_("The expiration date you entered is in the past.")])
            self._errors["expire_month"] = self.error_class([_("Check expiration")])
            return False
        return True


class DateRangeForm(forms.Form):
    start_date = forms.DateField(required=False, label=_("Start Date"), widget=SelectDateWidget())
    end_date = forms.DateField(required=False, label=_("End Date"), widget=SelectDateWidget())

    def clean(self):
        cleaned_data = super().clean()
        start_date = cleaned_data.get('start_date')
        end_date = cleaned_data.get('end_date')

        if end_date and end_date < start_date:
            self.add_error('end_date', _('End Date cannot be before Start Date'))
            del(cleaned_data['end_date'])
            
        return cleaned_data



import random
import string

from django.contrib.sites.models import Site
from django.utils.module_loading import import_string

from vendor.config import VENDOR_DATA_ENCODER, AVAILABLE_CURRENCIES

encode = import_string('{}.encode'.format(VENDOR_DATA_ENCODER))
decode = import_string('{}.decode'.format(VENDOR_DATA_ENCODER))

###########
# UTILITIES
###########

def random_string(length=8, check=[]):
    letters= string.digits + string.ascii_uppercase
    value = ''.join(random.sample(letters,length))

    if value not in check:
        return value
    return random_string(length=length, check=check)

def generate_sku():
    return random_string()

def set_default_site_id():
    return Site.objects.get_current()

def is_currency_available(msrp_currencies, currency=None):
    """
    Checks to see if the MSRP currencies for a product are available from the site's 
    configuration. If a currency is passed to the function, it will additionally check 
    if that currency is available from the filtered configured currencies and MSRP currencies.
    """
    available_currencies = set(AVAILABLE_CURRENCIES.keys()).intersection(msrp_currencies)

    if not currency and not available_currencies:
        return False
        
    if currency not in available_currencies:
        return False 
    
    return True



from django.shortcuts import render, redirect
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.sites.shortcuts import get_current_site
from django.urls import reverse, reverse_lazy
from django.conf import settings
from django.utils.translation import ugettext as _
from django.core.exceptions import ObjectDoesNotExist
from django.template import RequestContext

from django.views.generic.edit import DeleteView, UpdateView
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic import TemplateView, View, FormView

from iso4217 import Currency

from vendor.models import Offer, Invoice, Payment, Address, CustomerProfile, OrderItem, Receipt
from vendor.models.choice import TermType, PurchaseStatus
from vendor.models.utils import set_default_site_id
from vendor.processors import PaymentProcessor
from vendor.forms import BillingAddressForm, CreditCardForm, AccountInformationForm, AddressForm
# from vendor.models.address import Address as GoogleAddress

# The Payment Processor configured in settings.py
payment_processor = PaymentProcessor

# TODO: Need to remove the login required

def get_purchase_invoice(user):
    """
    Return an invoice that is in checkout or cart state or a newly create invoice in cart state.
    """
    profile, created = user.customer_profile.get_or_create(site=settings.SITE_ID)
    return profile.get_cart_or_checkout_cart()

def clear_session_purchase_data(request):
    if 'billing_address_form' in request.session:
        del(request.session['billing_address_form'])
    if 'credit_card_form' in request.session:
        del(request.session['credit_card_form'])

def get_or_create_session_cart(session):
    session_cart = {}
    if 'session_cart' not in session:
        session['session_cart'] = session_cart
    session_cart = session.get('session_cart')

    return session_cart
    
def check_offer_items_or_redirect(invoice, request):
    
    if invoice.order_items.count() < 1:
        messages.info(request, _("Please add to your cart"))
        redirect('vendor:cart')

class CartView(TemplateView):
    '''
    View items in the cart
    '''
    template_name = 'vendor/invoice_detail.html'

    def get(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        if request.user.is_anonymous:
            session_cart = get_or_create_session_cart(request.session)

            context['invoice'] = {}
            if len(session_cart):
                context['order_items'] = [ OrderItem(offer=Offer.objects.get(pk=offer), quantity=session_cart[offer]['quantity']) for offer in session_cart.keys() ]
            else:
                context['order_items'] = []
            context['invoice']['subtotal'] = sum([item.total for item in context['order_items'] ])
            context['invoice']['shipping'] = 0
            context['invoice']['tax'] = 0
            context['invoice']['total'] = context['invoice']['subtotal']

            return render(request, self.template_name, context)

        profile, created = self.request.user.customer_profile.get_or_create(site=set_default_site_id())
        cart = profile.get_cart_or_checkout_cart()
        context['invoice'] = cart
        context['order_items'] = [ order_item for order_item in cart.order_items.all() ]
        return render(request, self.template_name, context)


class AddToCartView(View):
    '''
    Create an order item and add it to the order
    '''
    def session_cart(self, request, offer):
        offer_key = str(offer.pk)
        session_cart = get_or_create_session_cart(request.session)

        if offer_key not in session_cart:
            session_cart[offer_key] = {}
            session_cart[offer_key]['quantity'] = 0

        session_cart[offer_key]['quantity'] += 1

        if not offer.allow_multiple:
            session_cart[offer_key]['quantity'] = 1

        return session_cart

    def post(self, request, *args, **kwargs):
        offer = Offer.on_site.get(slug=self.kwargs["slug"])
        if request.user.is_anonymous:
            request.session['session_cart'] = self.session_cart(request, offer)
        else:
            profile, created = self.request.user.customer_profile.get_or_create(site=get_current_site(request))

            cart = profile.get_cart_or_checkout_cart()

            if cart.status == Invoice.InvoiceStatus.CHECKOUT:
                cart.status = Invoice.InvoiceStatus.CART
                cart.save()

            if profile.has_product(offer.products.all()) and not offer.allow_multiple:
                messages.info(self.request, _("You Have Already Purchased This Item"))
            elif cart.order_items.filter(offer__products__in=offer.products.all()).count() and not offer.allow_multiple:
                messages.info(self.request, _("You already have this product in you cart. You can only buy one"))
            else:
                messages.info(self.request, _("Added item to cart."))
                cart.add_offer(offer)

        return redirect('vendor:cart')      # Redirect to cart on success


class RemoveFromCartView(View):
    
    def post(self, request, *args, **kwargs):
        offer = Offer.on_site.get(slug=self.kwargs["slug"])
        if request.user.is_anonymous:
            offer_key = str(offer.pk)
            session_cart = get_or_create_session_cart(request.session)

            if offer_key in session_cart:
                session_cart[offer_key]['quantity'] -= 1

            if session_cart[offer_key]['quantity'] <= 0:
                del(session_cart[offer_key])

            request.session['session_cart'] = session_cart
        else:
            profile = self.request.user.customer_profile.get(site=get_current_site(self.request))      # Make sure they have a cart

            cart = profile.get_cart_or_checkout_cart()

            if cart.status == Invoice.InvoiceStatus.CHECKOUT:
                cart.status = Invoice.InvoiceStatus.CART
                cart.save()

            cart.remove_offer(offer)

        messages.info(self.request, _("Removed item from cart."))

        return redirect('vendor:cart')      # Redirect to cart on success


class AccountInformationView(LoginRequiredMixin, TemplateView):
    template_name = 'vendor/checkout.html'

    def get(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        clear_session_purchase_data(request)
        
        invoice = get_purchase_invoice(request.user)
        if not invoice.order_items.count():
            return redirect('vendor:cart')
        
        invoice.status = Invoice.InvoiceStatus.CHECKOUT
        invoice.save()

        existing_account_address = Address.objects.filter(profile__user=request.user)

        if existing_account_address:
            # TODO: In future the user will be able to select from multiple saved address
            form = AccountInformationForm(initial={'email': request.user.email}, instance=existing_account_address[0])
        else:
            form = AccountInformationForm(initial={'first_name': request.user.first_name, 'last_name': request.user.last_name, 'email': request.user.email})
        
        context['form'] = form
        context['invoice'] = invoice
        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        form = AccountInformationForm(request.POST)

        if not form.is_valid():
            return render(request, self.template_name, {'form': form})

        shipping_address = form.save(commit=False)

        invoice = get_purchase_invoice(request.user)
        
        if not invoice.order_items.count() or invoice.status == Invoice.InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')

        invoice.status = Invoice.InvoiceStatus.CHECKOUT
        invoice.customer_notes = {'remittance_email': form.cleaned_data['email']}
        # TODO: Need to add a drop down to select existing address
        shipping_address, created = invoice.profile.get_or_create_address(shipping_address)
        if created:
            shipping_address.profile = invoice.profile
            shipping_address.save()
        invoice.shipping_address = shipping_address
        invoice.save()

        return redirect('vendor:checkout-payment')


class PaymentView(LoginRequiredMixin, TemplateView):
    '''
    Review items and submit Payment
    '''
    template_name = "vendor/checkout.html"

    def get(self, request, *args, **kwargs):
        invoice = get_purchase_invoice(request.user)
        if not invoice.order_items.count():
            return redirect('vendor:cart')

        context = super().get_context_data()

        processor = payment_processor(invoice)

        context = processor.get_checkout_context(context=context)

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        invoice = get_purchase_invoice(request.user)
        
        if not invoice.order_items.count() or invoice.status == Invoice.InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')

        credit_card_form = CreditCardForm(request.POST)
        if request.POST.get('same_as_shipping') == 'on':
            billing_address_form = BillingAddressForm(instance=invoice.shipping_address)
            billing_address_form.data = billing_address_form.initial
            billing_address_form.is_bound = True
        else:
            billing_address_form = BillingAddressForm(request.POST)

        if not (billing_address_form.is_valid() and credit_card_form.is_valid()):
            processor = payment_processor(invoice)
            context['billing_address_form'] = billing_address_form
            context['credit_card_form'] = credit_card_form
            return render(request, self.template_name, processor.get_checkout_context(context=context))
        else:
            billing_address_form.full_clean()
            credit_card_form.full_clean()
            request.session['billing_address_form'] = billing_address_form.cleaned_data
            request.session['credit_card_form'] = credit_card_form.cleaned_data
            return redirect('vendor:checkout-review')


class ReviewCheckoutView(LoginRequiredMixin, TemplateView):
    template_name = 'vendor/checkout.html'

    def get(self, request, *args, **kwargs):
        invoice = get_purchase_invoice(request.user)
        if not invoice.order_items.count():
            return redirect('vendor:cart')

        context = super().get_context_data()

        processor = payment_processor(invoice)
        if 'billing_address_form' in request.session:
            context['billing_address_form'] = BillingAddressForm(request.session['billing_address_form'])
        if 'credit_card_form' in request.session:
            context['credit_card_form'] = CreditCardForm(request.session['credit_card_form'])
        
        context = processor.get_checkout_context(context=context)
        
        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        invoice = get_purchase_invoice(request.user)
        
        if not invoice.order_items.count() or invoice.status == Invoice.InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')
        
        processor = payment_processor(invoice)
        
        processor.get_billing_address_form_data(request.session.get('billing_address_form'), BillingAddressForm)
        processor.get_payment_info_form_data(request.session.get('credit_card_form'), CreditCardForm)

        processor.authorize_payment()

        if processor.transaction_submitted:
            return redirect('vendor:purchase-summary', uuid=invoice.uuid)
        else:
            # TODO: Make message configurable for the site in the settings 
            messages.info(self.request, _("The payment gateway did not authorize payment."))
            return redirect('vendor:checkout-account')


class PaymentSummaryView(LoginRequiredMixin, DetailView):
    model = Invoice
    template_name = 'vendor/payment_summary.html'
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data()
        context['payment'] = self.object.payments.filter(success=True).first()
        return context


class OrderHistoryListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated by the current user on the current site.
    '''
    model = Invoice
    # TODO: filter to only include the current user's order history

    def get_queryset(self):
        try:
            # The profile and user are site specific so this should only return what's on the site for that user excluding the cart
            return self.request.user.customer_profile.get(site=get_current_site(self.request)).invoices.filter(status__gt=Invoice.InvoiceStatus.CART)
        except ObjectDoesNotExist:         # Catch the actual error for the exception
            return []   # Return empty list if there is no customer_profile


class OrderHistoryDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated by the current user on the current site.
    '''
    template_name = "vendor/invoice_history_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class ProductsListView(LoginRequiredMixin, ListView):
    model = Receipt
    template_name = 'vendor/purchase_list.html'

    def get_queryset(self):
        return self.request.user.customer_profile.get(site=get_current_site(self.request)).receipts.filter(status__gte=PurchaseStatus.COMPLETE)


class ReceiptDetailView(LoginRequiredMixin, DetailView):
    model = Receipt
    template_name = 'vendor/purchase_detail.html'
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)

        context['payment'] = self.object.order_item.invoice.payments.get(success=True, transaction=self.object.transaction)

        return context


class SubscriptionsListView(LoginRequiredMixin, ListView):
    model = Receipt
    template_name = 'vendor/purchase_list.html'

    def get_queryset(self):
        receipts = self.request.user.customer_profile.get(site=get_current_site(self.request)).receipts.filter(status__gte=PurchaseStatus.COMPLETE)
        subscriptions = [ receipt for receipt in receipts.all() if receipt.order_item.offer.terms > TermType.PERPETUAL and receipt.order_item.offer.terms < TermType.ONE_TIME_USE ]
        return subscriptions


class SubscriptionCancelView(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor:customer-subscriptions')

    def post(self, request, *args, **kwargs):
        receipt = Receipt.objects.get(uuid=self.kwargs["uuid"])

        processor = PaymentProcessor(receipt.order_item.invoice)

        processor.subscription_cancel(receipt)

        messages.info(self.request, _("Subscription Cancelled"))

        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class SubscriptionUpdatePaymentView(LoginRequiredMixin, FormView):
    form_class = CreditCardForm()
    success_url = reverse_lazy('vendor:customer-subscriptions')
    
    def post(self, request, *args, **kwargs):
        receipt = Receipt.objects.get(uuid=self.kwargs["uuid"])
        payment_form = CreditCardForm(request.POST)
        
        if not payment_form.is_valid():
            messages.info(request, _("Invalid Card"))
            return redirect(request.META.get('HTTP_REFERER', self.success_url))

        processor = PaymentProcessor(receipt.order_item.invoice)
        processor.get_payment_info_form_data(request.POST, CreditCardForm)
        processor.subscription_update_payment(receipt)

        if not processor.transaction_submitted:
            messages.info(request, _(f"Payment gateway error: {processor.transaction_message.get('message', '')}"))
            return redirect(request.META.get('HTTP_REFERER', self.success_url))
        
        messages.info(request, _(f"Success: Payment Updated"))
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class ShippingAddressUpdateView(LoginRequiredMixin, UpdateView):
    model = Address
    fields = '__all__'
    template_name_suffix = '_update_form'
    template_name = 'vendor/address_detail.html'
    success_url = reverse_lazy('vendor:products')

    def get_success_url(self):
        messages.info(self.request, _("Shipping Address Updated"))
        return self.request.META.get('HTTP_REFERER', self.success_url)
    




from django.urls import path

from vendor.views import vendor as vendor_views

app_name = "vendor"

urlpatterns = [
    path('cart/', vendor_views.CartView.as_view(), name="cart"),
    path('cart/add/<slug:slug>/', vendor_views.AddToCartView.as_view(), name="add-to-cart"),
    path('cart/remove/<slug:slug>/', vendor_views.RemoveFromCartView.as_view(), name="remove-from-cart"),
    path('checkout/summary/<uuid:uuid>/', vendor_views.PaymentSummaryView.as_view(), name="purchase-summary"),

    # path('cart/remove/<slug:slug>/', vendor_views.TransactionSummary.as_view(), name="transaction-summary"),
    # path('cart-item/edit/<int:id>/', vendor_views.CartItemQuantityEditView.as_view(), name='vendor-cart-item-quantity-edit'),
    # path('retrieve/cart/', vendor_views.RetrieveCartView.as_view(), name='vendor-user-cart-retrieve'),
    # path('delete/cart/<int:id>/', vendor_views.DeleteCartView.as_view(), name='vendor-user-cart-delete'),
    # path('retrieve/order/<int:id>/', vendor_views.RetrieveOrderView.as_view(), name='vendor-user-order-retrieve'),
    # path('retrieve/purchase-item/<int:id>/', vendor_views.RetrievePurchaseView.as_view(), name='vendor-user-purchase-retrieve'),
    # path('retrieve/purchase/list/', vendor_views.RetrievePurchaseListView.as_view(), name='vendor-user-purchase-list'),
    # path('retrieve/order-summary/', vendor_views.RetrieveOrderSummaryView.as_view(), name='vendor-order-summary-retrieve'),
    # path('payment/processing/', vendor_views.PaymentProcessingView.as_view(), name='vendor-payment-processing'),
    # path('request/refund/<int:id>/', vendor_views.RequestRefundView.as_view(), name='vendor-request-refund'),
    # path('retrieve/refund/requests/', vendor_views.RetrieveRefundRequestsView.as_view(), name='vendor-retrieve-refund-requests'),
    # path('issue/refund/<int:id>/', vendor_views.IssueRefundView.as_view(), name='vendor-issue-refund'),
    
    path('checkout/account/', vendor_views.AccountInformationView.as_view(), name="checkout-account"),
    path('checkout/payment/', vendor_views.PaymentView.as_view(), name="checkout-payment"),
    path('checkout/review/', vendor_views.ReviewCheckoutView.as_view(), name="checkout-review"),

    path('customer/products/', vendor_views.ProductsListView.as_view(), name="customer-products"),
    path('customer/product/<uuid:uuid>/receipt/', vendor_views.ReceiptDetailView.as_view(), name="customer-receipt"),
    path('customer/subscriptions/', vendor_views.SubscriptionsListView.as_view(), name="customer-subscriptions"),
    path('customer/subscription/<uuid:uuid>/cancel/', vendor_views.SubscriptionCancelView.as_view(), name="customer-subscription-cancel"),
    path('customer/subscription/update/<uuid:uuid>/payment', vendor_views.SubscriptionUpdatePaymentView.as_view(), name="customer-subscription-update-payment"),
    path('customer/shipping/<int:pk>/update', vendor_views.ShippingAddressUpdateView.as_view(), name="customer-shipping-update"),               # TODO: [GK-3030] Do not use PKs in URLs

    # TODO: Add user's account mangement urls
    # TODO: add user's order details page
]
